@@ -1,5 +1,5 @@
 /*
-** Copyright (C) 2022-2023 Dirk-Jan C. Binnema <djcb@djcbsoftware.nl>
+** Copyright (C) 2022-2024 Dirk-Jan C. Binnema <djcb@djcbsoftware.nl>
 **
 ** This program is free software; you can redistribute it and/or modify it
 ** under the terms of the GNU General Public License as published by the
@@ -16,7 +16,7 @@
 ** Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 **
 */
-
+#include "config.h"
 
 #include "mu-message.hh"
 #include "gmime/gmime-references.h"
@@ -29,6 +29,7 @@
 #include <utils/mu-utils.hh>
 #include <utils/mu-error.hh>
 #include <utils/mu-option.hh>
+#include <utils/mu-lang-detector.hh>
 
 #include <atomic>
 #include <mutex>
@@ -44,9 +45,10 @@
 using namespace Mu;
 
 struct Message::Private {
-	Private(Message::Options options): opts{options} {}
+	Private(Message::Options options):
+		opts{options}, doc{doc_opts(opts)} {}
 	Private(Message::Options options, Xapian::Document&& xdoc):
-		opts{options}, doc{std::move(xdoc)} {}
+		opts{options}, doc{std::move(xdoc), doc_opts(opts)} {}
 
 	Message::Options                opts;
 	Document			doc;
@@ -67,6 +69,15 @@ struct Message::Private {
 	Option<std::string> body_txt;
 	Option<std::string> body_html;
 	Option<std::string> embedded;
+
+	Option<std::string> language; /* body ISO language code */
+
+private:
+	Document::Options doc_opts(Message::Options mopts) {
+		return any_of(opts & Message::Options::SupportNgrams) ?
+			Document::Options::SupportNgrams :
+			Document::Options::None;
+	}
 };
 
 
@@ -77,20 +88,18 @@ get_statbuf(const std::string& path, Message::Options opts = Message::Options::N
 {
 	if (none_of(opts & Message::Options::AllowRelativePath) &&
 	    !g_path_is_absolute(path.c_str()))
-		return Err(Error::Code::File, "path '%s' is not absolute",
-			   path.c_str());
+		return Err(Error::Code::File, "path '{}' is not absolute", path);
 
 	if (::access(path.c_str(), R_OK) != 0)
-		return Err(Error::Code::File, "file @ '%s' is not readable",
-			   path.c_str());
+		return Err(Error::Code::File, "file @ '{}' is not readable", path);
 
 	struct stat statbuf{};
 	if (::stat(path.c_str(), &statbuf) < 0)
-		return Err(Error::Code::File, "cannot stat %s: %s", path.c_str(),
-			    g_strerror(errno));
+		return Err(Error::Code::File, "cannot stat {}: {}", path,
+			   g_strerror(errno));
 
 	if (!S_ISREG(statbuf.st_mode))
-		return Err(Error::Code::File, "not a regular file: %s", path.c_str());
+		return Err(Error::Code::File, "not a regular file: {}", path);
 
 	return Ok(std::move(statbuf));
 }
@@ -175,6 +184,11 @@ Message::document() const
 	return priv_->doc;
 }
 
+Message::Options
+Message::options() const
+{
+	return priv_->opts;
+}
 
 unsigned
 Message::docid() const
@@ -197,13 +211,13 @@ Message::set_maildir(const std::string& maildir)
 	    maildir.at(0) != '/' ||
 	    (maildir.size() > 1 && maildir.at(maildir.length()-1) == '/'))
 		return Err(Error::Code::Message,
-			   "'%s' is not a valid maildir", maildir.c_str());
+			   "'{}' is not a valid maildir", maildir.c_str());
 
 	const auto path{document().string_value(Field::Id::Path)};
 	if (path == maildir || path.find(maildir) == std::string::npos)
 		return Err(Error::Code::Message,
-			   "'%s' is not a valid maildir for message @ %s",
-			   maildir.c_str(), path.c_str());
+			   "'{}' is not a valid maildir for message @ {}",
+			   maildir, path);
 
 	priv_->doc.remove(Field::Id::Maildir);
 	priv_->doc.add(Field::Id::Maildir, maildir);
@@ -226,8 +240,8 @@ Message::load_mime_message(bool reload) const
 
 	const auto path{document().string_value(Field::Id::Path)};
 	if (auto mime_msg{MimeMessage::make_from_file(path)}; !mime_msg) {
-		g_warning("failed to load '%s': %s",
-			  path.c_str(), mime_msg.error().what());
+		mu_warning("failed to load '{}': {}",
+			   path, mime_msg.error().what());
 		return false;
 	} else {
 		priv_->mime_msg = std::move(mime_msg.value());
@@ -255,17 +269,17 @@ get_priority(const MimeMessage& mime_msg)
 	constexpr std::array<std::pair<std::string_view, Priority>, 10>
 		prio_alist = {{
 			{"high",	Priority::High},
-			{"1",		Priority::High},
-			{"2",		Priority::High},
+			{"1",	Priority::High},
+			{"2",	Priority::High},
 
 			{"normal",	Priority::Normal},
-			{"3",		Priority::Normal},
+			{"3",	Priority::Normal},
 
-			{"low",		Priority::Low},
+			{"low",	Priority::Low},
 			{"list",	Priority::Low},
 			{"bulk",	Priority::Low},
-			{"4",		Priority::Low},
-			{"5",		Priority::Low}
+			{"4",	Priority::Low},
+			{"5",	Priority::Low}
 		}};
 
 	const auto opt_str = mime_msg.header("Precedence")
@@ -312,8 +326,16 @@ get_mailing_list(const MimeMessage& mime_msg)
 	const char *b, *e;
 
 	const auto hdr{mime_msg.header("List-Id")};
-	if (!hdr)
-		return {};
+	if (!hdr) {
+		/* some marketing messages don't have a List-Id, but _do_ have a
+		 * List-Unsubscribe; if so, return an empty string here, so this
+		 * message is still flagged as "MailingList"
+		 */
+		if (const auto lu = mime_msg.header("List-Unsubscribe"); !!lu)
+			return "";
+		else
+			return Nothing;
+	}
 
 	dechdr = g_mime_utils_header_decode_phrase(NULL, hdr->c_str());
 	if (!dechdr)
@@ -335,11 +357,11 @@ get_mailing_list(const MimeMessage& mime_msg)
 }
 
 static void
-append_text(Option<std::string>& str, Option<std::string> app)
+append_text(Option<std::string>& str, Option<std::string>&& app)
 {
-	if (!str)
-		str = app;
-	else if (app)
+	if (!str && app)
+		str = std::move(*app);
+	else if (str && app)
 		str.value() += app.value();
 }
 
@@ -406,17 +428,18 @@ process_message_part(const MimeMessagePart& msg_part,
 		return;
 
 	submsg->for_each([&](auto&& parent, auto&& child_obj) {
-
-		/* XXX: we only handle one level */
-
+		/* NOTE: we only handle one level; ideally, we'd apply the whole
+		   parsing machinery recursively; so this a little crude. */
 		if (!child_obj.is_part())
 			return;
-
-		const auto ctype{child_obj.content_type()};
-		if (!ctype || !ctype->is_type("text", "*"))
+		if (const auto ctype{child_obj.content_type()}; !ctype)
 			return;
-
-		append_text(info.embedded, MimePart{child_obj}.to_string());
+		else if (ctype->is_type("text", "plain"))
+			append_text(info.embedded, MimePart{child_obj}.to_string());
+		else if (ctype->is_type("text", "html")) {
+			if (auto&& str{MimePart{child_obj}.to_string()}; str)
+				append_text(info.embedded, html_to_text(*str));
+		}
 	});
 }
 
@@ -437,14 +460,13 @@ handle_encrypted(const MimeMultipartEncrypted& part, Message::Private& info)
 	const auto proto{part.content_type_parameter("protocol").value_or("unknown")};
 	const auto ctx = MimeCryptoContext::make(proto);
 	if (!ctx) {
-		g_warning("failed to create context for protocol <%s>",
-			  proto.c_str());
+		mu_warning("failed to create context for protocol <{}>", proto);
 		return;
 	}
 
 	auto res{part.decrypt(*ctx)};
 	if (!res) {
-		g_warning("failed to decrypt: %s", res.error().what());
+		mu_warning("failed to decrypt: {}", res.error().what());
 		return;
 	}
 
@@ -531,6 +553,17 @@ process_message(const MimeMessage& mime_msg, const std::string& path,
 	info.mailing_list = get_mailing_list(mime_msg);
 	if (info.mailing_list)
 		info.flags |= Flags::MailingList;
+
+#ifdef HAVE_CLD2
+	/* language detection requires the cld2 lib */
+	if (info.body_txt) { /* attempt to get the body-language */
+		if (const auto lang{detect_language(info.body_txt.value())}; lang) {
+			info.language = lang->code;
+			mu_debug("detected language: {}", lang->code);
+		} else
+			mu_debug("could not detect language");
+	}
+#endif /*HAVE_CLD2*/
 }
 
 static Mu::Result<std::string>
@@ -540,8 +573,8 @@ calculate_sha256(const std::string& path)
 
 	FILE *file{::fopen(path.c_str(), "r")};
 	if (!file)
-		return Err(Error{Error::Code::File, "failed to open %s: %s",
-					path.c_str(), ::strerror(errno)});
+		return Err(Error{Error::Code::File, "failed to open {}: {}",
+					path, ::strerror(errno)});
 
 	std::array<uint8_t, 4096> buf{};
 	while (true) {
@@ -555,7 +588,7 @@ calculate_sha256(const std::string& path)
 	::fclose(file);
 
 	if (has_err)
-		return Err(Error{Error::Code::File, "failed to read %s", path.c_str()});
+		return Err(Error{Error::Code::File, "failed to read {}", path});
 
 	return Ok(g_checksum_get_string(checksum));
 }
@@ -574,20 +607,19 @@ fake_message_id(const std::string& path)
 
 	// not a very good message-id, only for testing.
 	if (path.empty() || ::access(path.c_str(), R_OK) != 0)
-		return format("%08x%s", g_str_hash(path.c_str()), mu_suffix);
+		return mu_format("{:08x}{}", g_str_hash(path.c_str()), mu_suffix);
 	if (const auto sha256_res{calculate_sha256(path)}; !sha256_res)
-		return format("%08x%s", g_str_hash(path.c_str()), mu_suffix);
+		return mu_format("{:08x}{}", g_str_hash(path.c_str()), mu_suffix);
 	else
-		return format("%s%s", sha256_res.value().c_str(), mu_suffix);
+		return mu_format("{}{}", sha256_res.value(), mu_suffix);
 }
 
-/* many of the doc.add(fiels ....) automatically update the sexp-list as well;
+/* many of the doc.add(fields ....) automatically update the sexp-list as well;
  * however, there are some _extra_ values in the sexp-list that are not
  * based on a field. So we add them here.
  */
 
 
-
 static void
 doc_add_list_post(Document& doc, const MimeMessage& mime_msg)
 {
@@ -643,7 +675,7 @@ fill_document(Message::Private& priv)
 	doc_add_reply_to(doc, mime_msg);  /* only in sexp */
 
 	field_for_each([&](auto&& field) {
-		/* insist on expliclity handling each */
+		/* insist on explicitly handling each */
 #pragma GCC diagnostic push
 #pragma GCC diagnostic error "-Wswitch"
 		switch(field.id) {
@@ -652,7 +684,8 @@ fill_document(Message::Private& priv)
 			break;
 		case Field::Id::BodyText:
 			doc.add(field.id, priv.body_txt);
-
+			if (priv.body_html)
+				doc.add(field.id, html_to_text(*priv.body_html));
 			break;
 		case Field::Id::Cc:
 			doc.add(field.id, mime_msg.contacts(Contact::Type::Cc));
@@ -676,6 +709,9 @@ fill_document(Message::Private& priv)
 		case Field::Id::From:
 			doc.add(field.id, mime_msg.contacts(Contact::Type::From));
 			break;
+		case Field::Id::Language:
+			doc.add(field.id, priv.language);
+			break;
 		case Field::Id::Maildir: /* already */
 			break;
 		case Field::Id::MailingList:
@@ -713,10 +749,6 @@ fill_document(Message::Private& priv)
 		case Field::Id::To:
 			doc.add(field.id, mime_msg.contacts(Contact::Type::To));
 			break;
-			/* internal fields */
-		case Field::Id::XBodyHtml:
-			doc.add(field.id, priv.body_html);
-			break;
 		/* LCOV_EXCL_START */
 		case Field::Id::_count_:
 		default:
@@ -786,11 +818,10 @@ Message::cache_path(Option<size_t> index) const
 
 	if (index) {
 		GError *err{};
-		auto tpath = format("%s/%zu", priv_->cache_path.c_str(), *index);
+		auto tpath = mu_format("{}/{}", priv_->cache_path, *index);
 		if (g_mkdir(tpath.c_str(), 0700) != 0)
 			return Err(Error::Code::File, &err,
-				   "failed to create cache dir '%s'; err=%d",
-				   tpath.c_str(), errno);
+				   "failed to create cache dir '{}'; err={}", tpath, errno);
 		return Ok(std::move(tpath));
 	} else
 
