@@ -1,5 +1,5 @@
 /*
-** Copyright (C) 2008-2021 Dirk-Jan C. Binnema <djcb@djcbsoftware.nl>
+** Copyright (C) 2008-2024 Dirk-Jan C. Binnema <djcb@djcbsoftware.nl>
 **
 ** This program is free software; you can redistribute it and/or modify
 ** it under the terms of the GNU General Public License as published by
@@ -26,21 +26,22 @@
 #include <cmath>
 
 #include <stdlib.h>
-#include <xapian.h>
 #include <glib/gstdio.h>
 
+#include "mu-xapian-db.hh"
 #include "mu-query-results.hh"
 #include "mu-query-match-deciders.hh"
 #include "mu-query-threads.hh"
-#include <mu-xapian.hh>
-#include "utils/mu-xapian-utils.hh"
+
+#include "mu-query-parser.hh"
 
 using namespace Mu;
 
 struct Query::Private {
-	Private(const Store& store) : store_{store}, parser_{store_} {}
-	// New
-	// bool calculate_threads (Xapian::Enquire& enq, size maxnum);
+	Private(const Store& store) :
+		store_{store},
+		parser_flags_{any_of(store_.message_options() & Message::Options::SupportNgrams) ?
+		ParserFlags::SupportNgrams : ParserFlags::None} {}
 
 	Xapian::Enquire make_enquire(const std::string& expr, Field::Id sortfield_id,
 				     QueryFlags qflags) const;
@@ -60,17 +61,12 @@ struct Query::Private {
 	Option<QueryResults> run(const std::string& expr,
 				 Field::Id sortfield_id, QueryFlags qflags,
 				 size_t maxnum) const;
-
-	size_t store_size() const { return store_.database().get_doccount(); }
-
 	const Store& store_;
-	const Parser parser_;
+	const ParserFlags parser_flags_;
 };
 
 Query::Query(const Store& store) : priv_{std::make_unique<Private>(store)} {}
 
-Query::Query(Query&& other) = default;
-
 Query::~Query() = default;
 
 static Xapian::Enquire&
@@ -82,24 +78,27 @@ sort_enquire(Xapian::Enquire& enq, Field::Id sortfield_id, QueryFlags qflags)
 	return enq;
 }
 
+static Xapian::Query
+make_query(const Store& store, const std::string& expr, ParserFlags parser_flags)
+{
+	if (expr.empty() || expr == R"("")")
+		return Xapian::Query::MatchAll;
+	else {
+		if (auto&& q{make_xapian_query(store, expr, parser_flags)}; !q) {
+			mu_warning("error in query '{}': {}", expr, q.error().what());
+			return Xapian::Query::MatchNothing;
+		} else
+			return q.value();
+	}
+}
+
 Xapian::Enquire
 Query::Private::make_enquire(const std::string& expr,
 			     Field::Id		sortfield_id,
 			     QueryFlags         qflags) const
 {
-	Xapian::Enquire enq{store_.database()};
-
-	if (expr.empty() || expr == R"("")")
-		enq.set_query(Xapian::Query::MatchAll);
-	else {
-		WarningVec warns;
-		const auto tree{parser_.parse(expr, warns)};
-		for (auto&& w : warns)
-			g_warning("query warning: %s", to_string(w).c_str());
-		enq.set_query(xapian_query(tree));
-		g_debug("qtree: %s", to_string(tree).c_str());
-	}
-
+	auto enq{store_.xapian_db().enquire()};
+	enq.set_query(make_query(store_, expr, parser_flags_));
 	sort_enquire(enq, sortfield_id, qflags);
 
 	return enq;
@@ -110,8 +109,10 @@ Query::Private::make_related_enquire(const StringSet& thread_ids,
 				     Field::Id sortfield_id,
 				     QueryFlags qflags) const
 {
-	Xapian::Enquire            enq{store_.database()};
+	auto enq{store_.xapian_db().enquire()};
 	std::vector<Xapian::Query> qvec;
+	qvec.reserve(thread_ids.size());
+
 	for (auto&& t : thread_ids)
 		qvec.emplace_back(field_from_id(Field::Id::ThreadId).xapian_term(t));
 
@@ -125,8 +126,7 @@ Query::Private::make_related_enquire(const StringSet& thread_ids,
 
 struct ThreadKeyMaker : public Xapian::KeyMaker {
 	ThreadKeyMaker(const QueryMatches& matches) : match_info_(matches) {}
-	std::string operator()(const Xapian::Document& doc) const override
-	{
+	std::string operator()(const Xapian::Document& doc) const override {
 		const auto it{match_info_.find(doc.get_docid())};
 		return (it == match_info_.end()) ? "" : it->second.thread_path;
 	}
@@ -216,11 +216,10 @@ Query::Private::run_related(const std::string& expr,
 
 	// Gather the thread-ids we found
 	mset.fetch();
-	for (auto it = mset.begin(); it != mset.end(); ++it) {
-		auto thread_id{opt_string(it.get_document(), Field::Id::ThreadId)};
-		if (thread_id)
+	minfo.thread_ids.reserve(mset.size());
+	for (auto it = mset.begin(); it != mset.end(); ++it)
+		if (auto thread_id{opt_string(it.get_document(), Field::Id::ThreadId)}; thread_id)
 			minfo.thread_ids.emplace(std::move(*thread_id));
-	}
 
 	// Now, determine the "related query".
 	//
@@ -230,12 +229,12 @@ Query::Private::run_related(const std::string& expr,
 	auto r_enq = std::invoke([&]{
 		if (threading)
 			return make_related_enquire(minfo.thread_ids, Field::Id::Date,
-						    qflags );
+						    qflags);
 		else
 			return make_related_enquire(minfo.thread_ids, sortfield_id, qflags);
 	});
 
-	const auto r_mset{r_enq.get_mset(0, threading ? store_size() : maxnum, {},
+	const auto r_mset{r_enq.get_mset(0, threading ? store_.size() : maxnum, {},
 					 make_related_decider(qflags, minfo).get())};
 	auto       qres{QueryResults{r_mset, std::move(minfo.matches)}};
 	return threading ? run_threaded(std::move(qres), r_enq, qflags, maxnum) : qres;
@@ -245,7 +244,7 @@ Option<QueryResults>
 Query::Private::run(const std::string&  expr, Field::Id sortfield_id, QueryFlags qflags,
 		    size_t maxnum) const
 {
-	const auto eff_maxnum{maxnum == 0 ? store_size() : maxnum};
+	const auto eff_maxnum{maxnum == 0 ? store_.size() : maxnum};
 
 	if (any_of(qflags & QueryFlags::IncludeRelated))
 		return run_related(expr, sortfield_id, qflags, eff_maxnum);
@@ -261,10 +260,13 @@ Query::run(const std::string& expr, Field::Id sortfield_id,
 	g_return_val_if_fail(none_of(qflags & QueryFlags::Leader),
 			     Err(Error::Code::InvalidArgument, "cannot pass Leader flag"));
 
-	StopWatch sw{format(
-	    "ran query '%s'; related: %s; threads: %s; max-size: %zu", expr.c_str(),
-	    any_of(qflags & QueryFlags::IncludeRelated) ? "yes" : "no",
-	    any_of(qflags & QueryFlags::Threading) ? "yes" : "no", maxnum)};
+	StopWatch sw{
+		mu_format("query: '{}'; (related:{}; threads:{}; ngrams:{}; max-size:{})",
+			  expr,
+			  any_of(qflags & QueryFlags::IncludeRelated) ? "yes" : "no",
+			  any_of(qflags & QueryFlags::Threading) ? "yes" : "no",
+			  any_of(priv_->parser_flags_ & ParserFlags::SupportNgrams) ? "yes" : "no",
+			  maxnum == 0 ? std::string{"∞"} : std::to_string(maxnum))};
 
 	return xapian_try_result([&]{
 		if (auto&& res = priv_->run(expr, sortfield_id, qflags, maxnum); res)
@@ -273,7 +275,6 @@ Query::run(const std::string& expr, Field::Id sortfield_id,
 			return Result<QueryResults>(Err(Error::Code::Query,
 							"failed to run query"));
 	});
-
 }
 
 size_t
@@ -282,7 +283,7 @@ Query::count(const std::string& expr) const
 	return xapian_try(
 	    [&] {
 		    const auto enq{priv_->make_enquire(expr, {}, {})};
-		    auto       mset{enq.get_mset(0, priv_->store_size())};
+		    auto       mset{enq.get_mset(0, priv_->store_.size())};
 		    mset.fetch();
 		    return mset.size();
 	    },
@@ -293,14 +294,10 @@ Query::count(const std::string& expr) const
 std::string
 Query::parse(const std::string& expr, bool xapian) const
 {
-	WarningVec warns;
-	const auto tree{priv_->parser_.parse(expr, warns)};
-	for (auto&& w : warns)
-		g_warning("query warning: %s", to_string(w).c_str());
-
 	if (xapian)
-		return xapian_query(tree).get_description();
+		return make_query(priv_->store_, expr,
+				  priv_->parser_flags_).get_description();
 	else
-		return to_string(tree);
+		return parse_query(expr).to_string();
 }
 /* LCOV_EXCL_STOP*/
