@@ -1,6 +1,6 @@
-;;; mu4e-contacts.el -- part of mu4e -*- lexical-binding: t -*-
+;;; mu4e-contacts.el --- Dealing with contacts -*- lexical-binding: t -*-
 
-;; Copyright (C) 2022 Dirk-Jan C. Binnema
+;; Copyright (C) 2022-2023 Dirk-Jan C. Binnema
 
 ;; Author: Dirk-Jan C. Binnema <djcb@djcbsoftware.nl>
 ;; Maintainer: Dirk-Jan C. Binnema <djcb@djcbsoftware.nl>
@@ -26,6 +26,7 @@
 
 ;;; Code:
 (require 'cl-lib)
+(require 'message)
 (require 'mu4e-helpers)
 (require 'mu4e-update)
 
@@ -79,13 +80,6 @@ their canonical counterpart; useful as an example."
         (mail (plist-get contact :mail)))
     (list :name name :mail mail)))
 
-(make-obsolete-variable 'mu4e-contact-rewrite-function
-                        "mu4e-contact-process-function (see docstring)"
-			"mu4e 1.3.2")
-(make-obsolete-variable 'mu4e-compose-complete-ignore-address-regexp
-                        "mu4e-contact-process-function (see docstring)"
-			"mu4e 1.3.2")
-
 (defcustom mu4e-contact-process-function
   (lambda(addr)
     (cond
@@ -129,48 +123,65 @@ predicate function. A value of nil keeps all the addresses."
   "Set with the full contact addresses for autocompletion.")
 
 ;;; user mail address
-(defun mu4e-personal-addresses(&optional no-regexp)
-  "Get the list user's personal addresses, as passed to mu init.
-The address are either plain e-mail address or /regular
- expressions/. When NO-REGEXP is non-nil, do not include regexp
+(defun mu4e-personal-addresses (&optional no-regexp)
+  "Get the list user's personal addresses, as passed to \"mu init\".
+
+The address are either plain e-mail addresses or regexps (strings
+ wrapped / /). When NO-REGEXP is non-nil, do not include regexp
  address patterns (if any)."
   (seq-remove
-   (lambda(addr) (and no-regexp (string-match-p "^/.*/" addr)))
-   (when (mu4e-server-properties)
-     (plist-get (mu4e-server-properties) :personal-addresses))))
+   (lambda (addr) (and no-regexp (string-match-p "^/.*/" addr)))
+   (when-let ((props (mu4e-server-properties)))
+     (plist-get props :personal-addresses))))
 
 (defun mu4e-personal-address-p (addr)
   "Is ADDR a personal address?
-Evaluate to nil if ADDR matches any of the personal addresses.
-Uses (mu4e-personal-addresses) for the addresses with both the plain
-addresses and /regular expressions/."
+Evaluate to nil if ADDR does not match any of the personal
+addresses.  Uses \\=(mu4e-personal-addresses) for the addresses
+with both the plain addresses and /regular expressions/."
   (when addr
     (seq-find
      (lambda (m)
        (if (string-match "/\\(.*\\)/" m)
            (let ((rx (match-string 1 m))
                  (case-fold-search t))
-             (if (string-match rx addr) t nil))
+             (string-match rx addr))
          (eq t (compare-strings addr nil nil m nil nil 'case-insensitive))))
      (mu4e-personal-addresses))))
 
-(define-obsolete-function-alias 'mu4e-user-mail-address-p
-  'mu4e-personal-address-p "1.5.5")
 
+(defun mu4e-personal-or-alternative-address-p (addr)
+  "Is ADDR either a personal or an alternative address?
 
-;; don't use the older vars anymore
-(make-obsolete-variable 'mu4e-user-mail-address-regexp
-                        'mu4e-user-mail-address-list "0.9.9.x")
-(make-obsolete-variable 'mu4e-my-email-addresses
-                        'mu4e-user-mail-address-list "0.9.9.x")
-(make-obsolete-variable 'mu4e-user-mail-address-list
-                        "determined by server; see `mu4e-personal-addresses'."
-			"1.3.8")
+That is, does it match either `mu4e-personal-address-p' or
+`message-alternative-emails'.
+
+Note that this expanded definition of user-addresses is only used
+in emacs, not in `mu' (e.g. when indexing).
+
+Also see `mu4e-personal-or-alternative-address-or-empty-p'."
+  (let ((alts message-alternative-emails))
+    (or (mu4e-personal-address-p addr)
+        (cond
+         ((functionp alts) (funcall alts addr))
+         ((stringp alts)   (string-match alts addr))
+         (t nil)))))
+
+(defun mu4e-personal-or-alternative-address-or-empty-p (addr)
+  "Is ADDR either a personal, alternative address or nil?
+
+This is like `mu4e-personal-or-alternative-address-p' but also
+return t for _empty_ ADDR. This can be useful for use with
+`message-dont-reply-to-names' since it can receive empty strings;
+those can be filtered-out by returning t here.
+
+See #2680 for further details. "
+  (or (and addr (string= addr ""))
+      (mu4e-personal-or-alternative-address-p addr)))
 
 
 ;; Helpers
 
-
 ;;; RFC2822 handling of phrases in mail-addresses
 ;;
 ;; The optional display-name contains a phrase, it sits before the
@@ -192,7 +203,7 @@ matches, nil is returned, if not, it returns a symbol
    ((= (aref ph 0) ?\")
     (if (string-match "\"\\([^\"\\\n]\\|\\\\.\\|\\\\\n\\)*\"" ph)
         'rfc822-quoted-string
-      'rfc822-containing-quote)) ; starts with quote, but doesn't end with one
+      'rfc822-containing-quote))   ; starts with quote, but doesn't end with one
    ((string-match-p "[\"]" ph) 'rfc822-containing-quote)
    ((string-match-p "[\000-\037()\*<>@,;:\\\.]+" ph) nil)
    (t 'rfc822-atom)))
@@ -225,15 +236,15 @@ case a phrase contains a quote, it will be escaped."
      (mu4e-contact-email contact)))
 
 (defsubst mu4e-contact-make (name email)
-  "Creata contact plist from NAME and EMAIL."
+  "Create a contact plist from NAME and EMAIL."
     `(:name ,name :email ,email))
 
 (defun mu4e-contact-full (contact)
   "Get the full combination of name and email address from CONTACT."
   (let* ((email (mu4e-contact-email contact))
-	 (name (mu4e-contact-name contact)))
+         (name (mu4e-contact-name contact)))
     (if (and name (> (length name) 0))
-	(format "%s <%s>" (mu4e--rfc822-quote-phrase name) email)
+        (format "%s <%s>" (mu4e--rfc822-quote-phrase name) email)
       email)))
 
 
@@ -249,9 +260,9 @@ This is used by the completion function in mu4e-compose."
     (dolist (contact contacts)
       (cl-incf n)
       (when (functionp mu4e-contact-process-function)
-	(setq contact (funcall mu4e-contact-process-function contact)))
+        (setq contact (funcall mu4e-contact-process-function contact)))
       (when contact ;; note the explicit deccode; the strings we get are
-		      ;; utf-8, but emacs doesn't know yet.
+                      ;; utf-8, but emacs doesn't know yet.
         (puthash (decode-coding-string contact 'utf-8) t mu4e--contacts-set)))
     (setq mu4e--contacts-tstamp (or tstamp "0"))
     (unless (zerop n)
@@ -265,7 +276,7 @@ For testing/debugging."
   (with-current-buffer (get-buffer-create "*mu4e-contacts-info*")
     (erase-buffer)
     (insert (format "complete addresses:        %s\n"
-		    (if mu4e-compose-complete-addresses "yes" "no")))
+                    (if mu4e-compose-complete-addresses "yes" "no")))
     (insert (format "only personal addresses:   %s\n"
                     (if mu4e-compose-complete-only-personal "yes" "no")))
     (insert (format "only addresses seen after: %s\n"
@@ -275,13 +286,13 @@ For testing/debugging."
       (insert (format "number of contacts cached: %d\n\n"
                       (hash-table-count mu4e--contacts-set)))
       (maphash (lambda (contact _)
-		 (insert (format "%s\n" contact))) mu4e--contacts-set))
+                 (insert (format "%s\n" contact))) mu4e--contacts-set))
     (pop-to-buffer "*mu4e-contacts-info*")))
 
 (declare-function mu4e--server-contacts  "mu4e-server")
 
 (defun mu4e--request-contacts-maybe ()
-  "Maybe update the set of contacts for autocompletion. 
+  "Maybe update the set of contacts for autocompletion.
 
 If `mu4e-compose-complete-addresses' is non-nil, get/update the
 list of contacts we use for autocompletion; otherwise, do
