@@ -1,24 +1,24 @@
-;;; mu4e-mark.el -- part of mu4e, the mu mail user agent -*- lexical-binding: t -*-
+;;; mu4e-mark.el --- Marking messages -*- lexical-binding: t -*-
 
-;; Copyright (C) 2011-2020 Dirk-Jan C. Binnema
+;; Copyright (C) 2011-2024 Dirk-Jan C. Binnema
 
 ;; Author: Dirk-Jan C. Binnema <djcb@djcbsoftware.nl>
 ;; Maintainer: Dirk-Jan C. Binnema <djcb@djcbsoftware.nl>
 
 ;; This file is not part of GNU Emacs.
 
-;; GNU Emacs is free software: you can redistribute it and/or modify
+;; mu4e is free software: you can redistribute it and/or modify
 ;; it under the terms of the GNU General Public License as published by
 ;; the Free Software Foundation, either version 3 of the License, or
 ;; (at your option) any later version.
 
-;; GNU Emacs is distributed in the hope that it will be useful,
+;; mu4e is distributed in the hope that it will be useful,
 ;; but WITHOUT ANY WARRANTY; without even the implied warranty of
 ;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 ;; GNU General Public License for more details.
 
 ;; You should have received a copy of the GNU General Public License
-;; along with GNU Emacs.  If not, see <http://www.gnu.org/licenses/>.
+;; along with mu4e.  If not, see <http://www.gnu.org/licenses/>.
 
 ;;; Commentary:
 
@@ -27,10 +27,9 @@
 
 ;;; Code:
 
-(require 'cl-lib)
-(require 'mu4e-proc)
-(require 'mu4e-utils)
+(require 'mu4e-server)
 (require 'mu4e-message)
+(require 'mu4e-folders)
 
 ;; keep byte-compiler happy
 (declare-function mu4e~headers-mark "mu4e-headers")
@@ -40,15 +39,18 @@
 ;;; Variables & constants
 
 (defcustom mu4e-headers-leave-behavior 'ask
-  "What to do when user leaves the headers view.
-That is when he e.g. quits, refreshes or does a new search.
+  "What to do when user leaves the current headers view.
+
+\"Leaving\" here means quitting the headers views, refreshing it
+or even quitting mu4e or Emacs.
+
 Value is one of the following symbols:
 - `ask'     ask user whether to ignore the marks
 - `apply'   automatically apply the marks before doing anything else
 - `ignore'  automatically ignore the marks without asking"
-  :type '(choice (const ask    :tag "ask user whether to ignore marks")
-                 (const apply  :tag "apply marks without asking")
-                 (const ignore :tag "ignore marks without asking"))
+  :type '(choice (const :tag "ask user whether to ignore marks" ask)
+                 (const :tag "apply marks without asking"       apply)
+                 (const :tag "ignore marks without asking"      ignore))
   :group 'mu4e-headers)
 
 (defcustom mu4e-mark-execute-pre-hook nil
@@ -59,21 +61,21 @@ executed and the message itself."
   :group 'mu4e-headers)
 
 (defvar mu4e-headers-show-target t
-  "Whether to show targets (such as '-> delete', '-> /archive')
-when marking message. Normally, this is useful information for the
-user, however, when you often mark large numbers (thousands) of
-message, showing the target makes this quite a bit slower (showing
-the target uses an Emacs feature called 'overlays', which aren't
-particularly fast).")
+  "Whether to show targets (such as \"-> delete\", \"-> /archive\")
+when marking message. Normally, this is useful information for
+the user, however, when you often mark large numbers (thousands)
+of message, showing the target makes this quite a bit
+slower (showing the target uses Emacs overlays, which can be slow
+when overused).")
 
 ;;; Insert stuff
 
-(defvar mu4e~mark-map nil
+(defvar mu4e--mark-map nil
   "Contains a mapping of docid->markinfo.
 When a message is marked, the information is added here. markinfo
-is a cons cell consisting of the following: \(mark . target)
-where MARK is the type of mark (move, trash, delete)
-TARGET (optional) is the target directory (for 'move')")
+is a cons cell consisting of the following: (mark . target) where
+MARK is the type of mark (move, trash, delete) TARGET (optional)
+is the target directory (for \"move\")")
 
 ;; the mark-map is specific for the current header buffer
 ;; currently, there can't be more than one, but we never know what will
@@ -82,49 +84,46 @@ TARGET (optional) is the target directory (for 'move')")
 ;; the fringe is the space on the left of headers, where we put marks below some
 ;; handy definitions; only `mu4e-mark-fringe-len' should be change (if ever),
 ;; the others follow from that.
-(defconst mu4e~mark-fringe-len 2
+(defconst mu4e--mark-fringe-len 2
   "Width of the fringe for marks on the left.")
-(defconst mu4e~mark-fringe (make-string mu4e~mark-fringe-len ?\s)
+(defconst mu4e--mark-fringe (make-string mu4e--mark-fringe-len ?\s)
   "The space on the left of message headers to put marks.")
-(defconst mu4e~mark-fringe-format (format "%%-%ds" mu4e~mark-fringe-len)
+(defconst mu4e--mark-fringe-format (format "%%-%ds" mu4e--mark-fringe-len)
   "Format string to set a mark and leave remaining space.")
 
-(defun mu4e~mark-initialize ()
+(defun mu4e--mark-initialize ()
   "Initialize the marks-subsystem."
-  (set (make-local-variable 'mu4e~mark-map) (make-hash-table)))
+  (set (make-local-variable 'mu4e--mark-map) (make-hash-table))
+  ;; ask user when kill buffer / emacs with live marks.
+  ;; (subject to mu4e-headers-leave-behavior)
+  (add-hook 'kill-buffer-query-functions
+            #'mu4e-mark-handle-when-leaving nil t)
+  (add-hook 'kill-emacs-query-functions
+            #'mu4e-mark-handle-when-leaving nil t))
 
-(defun mu4e~mark-clear ()
+(defun mu4e--mark-clear ()
   "Clear the marks-subsystem."
-  (clrhash mu4e~mark-map))
+  (clrhash mu4e--mark-map))
 
-(defun mu4e~mark-find-headers-buffer ()
+(defun mu4e--mark-find-headers-buffer ()
   "Find the headers buffer, if any."
-  (cl-find-if
-   (lambda (b)
-     (with-current-buffer b
-       (eq major-mode 'mu4e-headers-mode)))
-   (buffer-list)))
+  (seq-find (lambda (_)
+              (mu4e-current-buffer-type-p 'headers))
+            (buffer-list)))
 
-(defmacro mu4e~mark-in-context (&rest body)
+(defmacro mu4e--mark-in-context (&rest body)
   "Evaluate BODY in the context of the headers buffer.
 The current buffer must be either a headers or view buffer."
   `(cond
-    ((eq major-mode 'mu4e-headers-mode) ,@body)
-    ((eq major-mode 'mu4e-view-mode)
+    ((mu4e-current-buffer-type-p 'headers) ,@body)
+    ((mu4e-current-buffer-type-p 'view)
      (when (buffer-live-p (mu4e-get-headers-buffer))
        (let* ((msg (mu4e-message-at-point))
               (docid (mu4e-message-field msg :docid)))
          (with-current-buffer (mu4e-get-headers-buffer)
-           (if (mu4e~headers-goto-docid docid)
-               ,@body
-             (mu4e-error "Cannot find message in headers buffer"))))))
-    (t
-     ;; even in other modes (e.g. mu4e-main-mode we try to find
-     ;; the headers buffer
-     (let ((hbuf (mu4e~mark-find-headers-buffer)))
-       (if (buffer-live-p hbuf)
-           (with-current-buffer hbuf ,@body)
-         ,@body)))))
+           (when (mu4e~headers-goto-docid docid)
+             ,@body
+             )))))))
 
 (defconst mu4e-marks
   '((refile
@@ -132,50 +131,51 @@ The current buffer must be either a headers or view buffer."
      :prompt "refile"
      :dyn-target (lambda (target msg) (mu4e-get-refile-folder msg))
      :action (lambda (docid msg target)
-               (mu4e~proc-move docid (mu4e~mark-check-target target) "-N")))
+               (mu4e--server-move docid (mu4e--mark-check-target target) "-N")))
     (delete
      :char ("D" . "x")
      :prompt "Delete"
      :show-target (lambda (target) "delete")
-     :action (lambda (docid msg target) (mu4e~proc-remove docid)))
+     :action (lambda (docid msg target) (mu4e--server-remove docid)))
     (flag
      :char ("+" . "✚")
      :prompt "+flag"
      :show-target (lambda (target) "flag")
      :action (lambda (docid msg target)
-               (mu4e~proc-move docid nil "+F-u-N")))
+               (mu4e--server-move docid nil "+F-u-N")))
     (move
      :char ("m" . "▷")
      :prompt "move"
-     :ask-target  mu4e~mark-get-move-target
+     :ask-target  mu4e--mark-get-move-target
      :action (lambda (docid msg target)
-               (mu4e~proc-move docid (mu4e~mark-check-target target) "-N")))
+               (mu4e--server-move docid (mu4e--mark-check-target target) "-N")))
     (read
      :char    ("!" . "◼")
      :prompt "!read"
      :show-target (lambda (target) "read")
-     :action (lambda (docid msg target) (mu4e~proc-move docid nil "+S-u-N")))
+     :action (lambda (docid msg target) (mu4e--server-move docid nil "+S-u-N")))
     (trash
      :char ("d" . "▼")
      :prompt "dtrash"
      :dyn-target (lambda (target msg) (mu4e-get-trash-folder msg))
-     :action (lambda (docid msg target) (mu4e~proc-move docid
-                                                        (mu4e~mark-check-target target) "+T-N")))
+     :action (lambda (docid msg target)
+               (mu4e--server-move docid
+                                  (mu4e--mark-check-target target) "+T-N")))
     (unflag
      :char    ("-" . "➖")
      :prompt "-unflag"
      :show-target (lambda (target) "unflag")
-     :action (lambda (docid msg target) (mu4e~proc-move docid nil "-F-N")))
+     :action (lambda (docid msg target) (mu4e--server-move docid nil "-F-N")))
     (untrash
      :char   ("=" . "▲")
      :prompt "=untrash"
      :show-target (lambda (target) "untrash")
-     :action (lambda (docid msg target) (mu4e~proc-move docid nil "-T")))
+     :action (lambda (docid msg target) (mu4e--server-move docid nil "-T")))
     (unread
      :char    ("?" . "◻")
      :prompt "?unread"
      :show-target (lambda (target) "unread")
-     :action (lambda (docid msg target) (mu4e~proc-move docid nil "-S+u-N")))
+     :action (lambda (docid msg target) (mu4e--server-move docid nil "-S+u-N")))
     (unmark
      :char  " "
      :prompt "unmark"
@@ -219,9 +219,9 @@ properties are:
 (defun mu4e-mark-at-point (mark target)
   "Mark (or unmark) message at point.
 MARK specifies the mark-type. For `move'-marks and `trash'-marks
-the TARGET argument is non-nil and specifies to which
-maildir the message is to be moved/trashed. The function works in
-both headers buffers and message buffers.
+the TARGET argument is non-nil and specifies to which maildir the
+message is to be moved/trashed. The function works in both
+headers buffers and message buffers.
 
 The following marks are available, and the corresponding props:
 
@@ -235,16 +235,16 @@ The following marks are available, and the corresponding props:
    `read'      n        mark the message as read
    `trash'     y        trash the message to some folder
    `unflag'    n        mark this message for unflagging
-   `untrash'   n        remove the 'trashed' flag from a message
+   `untrash'   n        remove the `trashed' flag from a message
    `unmark'    n        unmark this message
    `unread'    n        mark the message as unread
    `action'    y        mark the message for some action."
   (interactive)
   (let* ((msg (mu4e-message-at-point))
          (docid (mu4e-message-field msg :docid))
-         ;; get a cell with the mark char and the 'target' 'move' already has a
-         ;; target (the target folder) the other ones get a pseudo "target", as
-         ;; info for the user.
+         ;; get a cell with the mark char and the "move" already has a target
+         ;; (the target folder) the other ones get a pseudo "target", as info
+         ;; for the user.
          (markdesc (cdr (or (assq mark mu4e-marks)
                             (mu4e-error "Invalid mark %S" mark))))
          (get-markkar
@@ -253,7 +253,7 @@ The following marks are available, and the corresponding props:
                 (if mu4e-use-fancy-chars (cdr char) (car char))
               char)))
          (markkar (funcall get-markkar (plist-get markdesc :char)))
-         (target (mu4e~mark-get-dyn-target mark target))
+         (target (mu4e--mark-get-dyn-target mark target))
          (show-fct (plist-get markdesc :show-target))
          (shown-target (if show-fct
                            (funcall show-fct target)
@@ -265,46 +265,44 @@ The following marks are available, and the corresponding props:
       (when (mu4e~headers-mark docid markkar)
         ;; update the hash -- remove everything current, and if add the new
         ;; stuff, unless we're unmarking
-        (remhash docid mu4e~mark-map)
-        ;; remove possible overlays
-        (remove-overlays (line-beginning-position) (line-end-position))
+        (remhash docid mu4e--mark-map)
+        ;; remove possible mark overlays
+        (remove-overlays (line-beginning-position) (line-end-position)
+                         'mu4e-mark t)
         ;; now, let's set a mark (unless we were unmarking)
         (unless (eql mark 'unmark)
-          (puthash docid (cons mark target) mu4e~mark-map)
+          (puthash docid (cons mark target) mu4e--mark-map)
           ;; when we have a target (ie., when moving), show the target folder in
           ;; an overlay
           (when (and shown-target mu4e-headers-show-target)
             (let* ((targetstr (propertize (concat "-> " shown-target " ")
                                           'face 'mu4e-system-face))
                    ;; mu4e~headers-goto-docid docid t \will take us just after
-                   ;; the docid cookie and then we skip the mu4e~mark-fringe
-                   (start (+ (length mu4e~mark-fringe)
+                   ;; the docid cookie and then we skip the mu4e--mark-fringe
+                   (start (+ (length mu4e--mark-fringe)
                              (mu4e~headers-goto-docid docid t)))
                    (overlay (make-overlay start (+ start (length targetstr)))))
               (overlay-put overlay 'display targetstr)
+              (overlay-put overlay 'mu4e-mark t)
+              (overlay-put overlay 'evaporate t)
               docid)))))))
 
-(defun mu4e~mark-get-move-target ()
+(defun mu4e--mark-get-move-target ()
   "Ask for a move target, and propose to create it if it does not exist."
-  (interactive)
-  ;;  (mu4e-message-at-point) ;; raises error if there is none
   (let* ((target (mu4e-ask-maildir "Move message to: "))
          (target (if (string= (substring target 0 1) "/")
                      target
                    (concat "/" target)))
-         (fulltarget (concat (mu4e-root-maildir) target)))
-    (when (or (file-directory-p fulltarget)
-              (and (yes-or-no-p
-                    (format "%s does not exist.  Create now?" fulltarget))
-                   (mu4e~proc-mkdir fulltarget)))
+         (fulltarget (mu4e-join-paths (mu4e-root-maildir) target)))
+    (when (mu4e-create-maildir-maybe fulltarget)
       target)))
 
-(defun mu4e~mark-ask-target (mark)
+(defun mu4e--mark-ask-target (mark)
   "Ask the target for MARK, if the user should be asked the target."
   (let ((getter (plist-get (cdr (assq mark mu4e-marks)) :ask-target)))
     (and getter (funcall getter))))
 
-(defun mu4e~mark-get-dyn-target (mark target)
+(defun mu4e--mark-get-dyn-target (mark target)
   "Get the dynamic TARGET for MARK.
 The result may depend on the message at point."
   (let ((getter (plist-get (cdr (assq mark mu4e-marks)) :dyn-target)))
@@ -316,7 +314,7 @@ The result may depend on the message at point."
   "Mark the header at point with MARK or all in the region.
 Optionally, provide TARGET (for moves)."
   (unless target
-    (setq target (mu4e~mark-ask-target mark)))
+    (setq target (mu4e--mark-ask-target mark)))
   (if (not (use-region-p))
       ;; single message
       (mu4e-mark-at-point mark target)
@@ -330,15 +328,15 @@ Optionally, provide TARGET (for moves)."
 
 (defun mu4e-mark-restore (docid)
   "Restore the visual mark for the message with DOCID."
-  (let ((markcell (gethash docid mu4e~mark-map)))
+  (let ((markcell (gethash docid mu4e--mark-map)))
     (when markcell
       (save-excursion
         (when (mu4e~headers-goto-docid docid)
           (mu4e-mark-at-point (car markcell) (cdr markcell)))))))
 
-(defun mu4e~mark-get-markpair (prompt &optional allow-something)
+(defun mu4e--mark-get-markpair (prompt &optional allow-something)
   "Ask user with PROMPT for a mark and return (MARK . TARGET).
-If ALLOW-SOMETHING is non-nil, allow the 'something' pseudo mark
+If ALLOW-SOMETHING is non-nil, allow the `something' pseudo mark
 as well."
   (let* ((marks (mapcar (lambda (markdescr)
                           (cons (plist-get (cdr markdescr) :prompt)
@@ -346,9 +344,9 @@ as well."
                         mu4e-marks))
          (marks
           (if allow-something
-              marks (cl-remove-if (lambda (m) (eq 'something (cdr m))) marks)))
+              marks (seq-remove (lambda (m) (eq 'something (cdr m))) marks)))
          (mark (mu4e-read-option prompt marks))
-         (target (mu4e~mark-ask-target mark)))
+         (target (mu4e--mark-ask-target mark)))
     (cons mark target)))
 
 (defun mu4e-mark-resolve-deferred-marks ()
@@ -356,7 +354,7 @@ as well."
 If there are such marks, replace them with a _real_ mark (ask the
 user which one)."
   (interactive)
-  (mu4e~mark-in-context
+  (mu4e--mark-in-context
    (let ((markpair))
      (maphash
       (lambda (docid val)
@@ -364,15 +362,15 @@ user which one)."
           (when (eql mark 'something)
             (unless markpair
               (setq markpair
-                    (mu4e~mark-get-markpair "Set deferred mark(s) to: " nil)))
+                    (mu4e--mark-get-markpair "Set deferred mark(s) to: " nil)))
             (save-excursion
               (when (mu4e~headers-goto-docid docid)
                 (mu4e-mark-set (car markpair) (cdr markpair)))))))
-      mu4e~mark-map))))
+      mu4e--mark-map))))
 
-(defun mu4e~mark-check-target (target)
+(defun mu4e--mark-check-target (target)
   "Check if TARGET exists; if not, offer to create it."
-  (let ((fulltarget (concat (mu4e-root-maildir) target)))
+  (let ((fulltarget (mu4e-join-paths (mu4e-root-maildir) target)))
     (if (not (mu4e-create-maildir-maybe fulltarget))
         (mu4e-error "Target dir %s does not exist " fulltarget)
       target)))
@@ -389,9 +387,9 @@ flow. Therefore, we hide the message, which in practice seems to
 work well.
 
 If NO-CONFIRMATION is non-nil, don't ask user for confirmation."
-  (interactive)
-  (mu4e~mark-in-context
-   (let* ((marknum (hash-table-count mu4e~mark-map))
+  (interactive "P")
+  (mu4e--mark-in-context
+   (let* ((marknum (mu4e-mark-marks-num))
           (prompt (format "Are you sure you want to execute %d mark%s?"
                           marknum (if (> marknum 1) "s" ""))))
      (if (zerop marknum)
@@ -414,42 +412,46 @@ If NO-CONFIRMATION is non-nil, don't ask user for confirmation."
                     (funcall (plist-get (cdr markdescr) :action)
                              docid msg target))
                 (mu4e-error "Unrecognized mark %S" mark))))
-          mu4e~mark-map))
-       (mu4e-mark-unmark-all)
+          mu4e--mark-map))
+       (mu4e-mark-unmark-all 'no-confirm)
        (message nil)))))
 
-(defun mu4e-mark-unmark-all ()
+(defun mu4e-mark-unmark-all (&optional no-confirmation)
   "Unmark all marked messages."
   (interactive)
-  (mu4e~mark-in-context
-   (when (or (null mu4e~mark-map) (zerop (hash-table-count mu4e~mark-map)))
+  (mu4e--mark-in-context
+   (when (zerop (mu4e-mark-marks-num))
      (mu4e-warn "Nothing is marked"))
-   (maphash
-    (lambda (docid _val)
-      (save-excursion
-        (when (mu4e~headers-goto-docid docid)
-          (mu4e-mark-set 'unmark))))
-    mu4e~mark-map)
-   ;; in any case, clear the marks map
-   (mu4e~mark-clear)))
+   (let* ((marknum (hash-table-count mu4e--mark-map))
+          (prompt (format "Are you sure you want to unmark %d message%s?"
+                          marknum (if (> marknum 1) "s" ""))))
+     (when (or no-confirmation (y-or-n-p prompt))
+       (maphash
+        (lambda (docid _val)
+          (save-excursion
+            (when (mu4e~headers-goto-docid docid)
+              (mu4e-mark-set 'unmark))))
+        mu4e--mark-map)
+       ;; in any case, clear the marks map
+       (mu4e--mark-clear)))))
 
 (defun mu4e-mark-docid-marked-p (docid)
   "Is the given DOCID marked?"
-  (when (gethash docid mu4e~mark-map) t))
-
+  (when (gethash docid mu4e--mark-map) t))
 
 (defun mu4e-mark-marks-num ()
   "Return the number of mark-instances in the current buffer."
-  (mu4e~mark-in-context
-   (if mu4e~mark-map (hash-table-count mu4e~mark-map) 0)))
+  (mu4e--mark-in-context
+   (if mu4e--mark-map (hash-table-count mu4e--mark-map) 0)))
 
 (defun mu4e-mark-handle-when-leaving ()
   "Handle any mark-instances in the current buffer when leaving.
-This is done according to the value of `mu4e-headers-leave-behavior'. This
-function is to be called before any further action (like searching,
-quitting the buffer) is taken; returning t means 'take the following
-action', return nil means 'don't do anything'."
-  (mu4e~mark-in-context
+This is done according to the value of
+`mu4e-headers-leave-behavior'. This function is to be called
+before any further action (like searching, quitting the buffer)
+is taken; returning t means \"take the following action\", return
+nil means \"don't do anything\"."
+  (mu4e--mark-in-context
    (let ((marknum (mu4e-mark-marks-num))
          (what mu4e-headers-leave-behavior))
      (unless (zerop marknum) ;; nothing to do?
@@ -461,7 +463,8 @@ action', return nil means 'don't do anything'."
                         ("ignore marks?" . ignore)))))
        ;; we determined what to do... now do it
        (when (eq what 'apply)
-         (mu4e-mark-execute-all t))))))
+         (mu4e-mark-execute-all t)))))
+  t) ;; return t for compat with `kill-buffer-query-functions
 
 ;;; _
 (provide 'mu4e-mark)
