@@ -1,6 +1,6 @@
-;;; mu4e-server.el -- part of mu4e -*- lexical-binding: t -*-
+;;; mu4e-server.el --- Control mu server from mu4e -*- lexical-binding: t -*-
 
-;; Copyright (C) 2011-2022 Dirk-Jan C. Binnema
+;; Copyright (C) 2011-2023 Dirk-Jan C. Binnema
 
 ;; Author: Dirk-Jan C. Binnema <djcb@djcbsoftware.nl>
 ;; Maintainer: Dirk-Jan C. Binnema <djcb@djcbsoftware.nl>
@@ -45,7 +45,7 @@ session."
 
 Changes to this value only take effect after (re)starting the mu
 session."
-  :type 'file
+  :type '(file :must-match t)
   :group 'mu4e
   :safe 'stringp)
 
@@ -58,10 +58,6 @@ session."
   :type 'boolean
   :group 'mu4e)
 
-(make-obsolete-variable
- 'mu4e-maildir
- "determined by server; see `mu4e-root-maildir'." "1.3.8")
-
 (defcustom mu4e-change-filenames-when-moving nil
   "Change message file names when moving them.
 
@@ -77,6 +73,30 @@ better with e.g. offlineimap."
   :group 'mu4e
   :safe 'booleanp)
 
+(defcustom mu4e-mu-allow-temp-file nil
+  "Allow using temp-files for optimizing mu <-> mu4e communication.
+
+Some commands - in particular \"find\" and \"contacts\" - return
+big s-expressions; and it turns out that reading those is faster
+by passing them through a temp file rather than through normal
+stdin/stdout channel - esp. on the (common case) where the
+file-system for temp-files is in-memory.
+
+To see if the helps, you can benchmark the rendering with
+     (setq mu4e-headers-report-render-time t)
+
+and compare the results with `mu4e-mu-allow-temp' set and unset.
+
+Note: for a change to this variable to take effect, you need to
+stop/start mu4e."
+  :type  'boolean
+  :group 'mu4e
+  :safe  'booleanp)
+
+
+;; Cached data
+(defvar mu4e-maildir-list)
+
 
 ;; Handlers are not strictly internal, but are not meant
 ;; for overriding outside mu4e. The are mainly for breaking
@@ -107,8 +127,6 @@ passed the docid and the draft-path of the sent message.")
 The function is passed a message sexp as argument. See
 `mu4e--server-filter' for the format.")
 
-(make-obsolete-variable 'mu4e-header-func "mu4e-headers-append-func" "1.7.4")
-
 (defvar mu4e-headers-append-func nil
   "Function called with a list of headers to append.
 The function is passed a list of message plists as argument. See
@@ -124,13 +142,6 @@ the number of matches. See `mu4e--server-filter' for the format.")
 This before new headers are displayed, to clear the current
 headers buffer. See `mu4e--server-filter' for the format.")
 
-(defvar mu4e-compose-func nil
-  "Function called for each compose message received.
-I.e., the original message that is used as basis for composing a
-new message (i.e., either a reply or a forward); the function is
-passed msg and a symbol (either reply or forward). See
-`mu4e--server-filter' for the format of <msg-plist>.")
-
 (defvar mu4e-info-func nil
   "Function called for each (:info type ....) sexp received.
 from the server process.")
@@ -138,13 +149,52 @@ from the server process.")
 (defvar mu4e-pong-func nil
   "Function called for each (:pong type ....) sexp received.")
 
+(defvar mu4e-queries-func nil
+  "Function called for each (:queries type ....) sexp received.")
+
 (defvar mu4e-contacts-func nil
-  "A function called for each (:contacts (<list-of-contacts>)
+  "A function called for each (:contacts (<list-of-contacts>))
 sexp received from the server process.")
 
-(make-obsolete-variable 'mu4e-temp-func "No longer used" "1.7.0")
 
-;;; Internal vars
+;;; Dealing with Server properties
+(defvar mu4e--server-props nil
+  "Metadata we receive from the mu4e server.")
+
+(defun mu4e-server-properties ()
+  "Get the server metadata plist."
+  mu4e--server-props)
+
+(defun mu4e-root-maildir()
+  "Get the root maildir."
+  (or (and mu4e--server-props
+           (plist-get mu4e--server-props :root-maildir))
+      (mu4e-error "Root maildir unknown; did you start mu4e?")))
+
+(defun mu4e-database-path()
+  "Get the root maildir."
+  (or (and mu4e--server-props
+           (plist-get mu4e--server-props :database-path))
+      (mu4e-error "Root maildir unknown; did you start mu4e?")))
+
+(defun mu4e-server-version()
+  "Get the root maildir."
+  (or (and mu4e--server-props
+           (plist-get mu4e--server-props :version))
+      (mu4e-error "Version unknown; did you start mu4e?")))
+
+;;; remember queries result.
+(defvar mu4e--server-query-items nil
+  "Query items results we receive from the mu4e server.
+Those are the results from the counting-queries
+for bookmarks and maildirs.")
+
+(defun mu4e-server-query-items ()
+  "Get the latest server query items."
+  mu4e--server-query-items)
+
+
+;;; Handling raw server data
 
 (defvar mu4e--server-buf nil
   "Buffer (string) for data received from the backend.")
@@ -162,7 +212,7 @@ sexp received from the server process.")
 <`mu4e--server-cookie-pre'><length-in-hex><`mu4e--server-cookie-post'>.")
 (defconst mu4e--server-cookie-matcher-rx
   (concat mu4e--server-cookie-pre "\\([[:xdigit:]]+\\)"
-	  mu4e--server-cookie-post)
+          mu4e--server-cookie-post)
   "Regular expression matching the length cookie.
 Match 1 will be the length (in hex).")
 
@@ -171,11 +221,10 @@ Match 1 will be the length (in hex).")
 Checks whether the server process is live."
   (and mu4e--server-process
        (memq (process-status mu4e--server-process)
-             '(run open listen connect stop))
-       t))
+             '(run open listen connect stop)) t))
 
 (defsubst mu4e--server-eat-sexp-from-buf ()
-  "'Eat' the next s-expression from `mu4e--server-buf'.
+  "Eat the next s-expression from `mu4e--server-buf'.
 Note: this is a string, not an emacs-buffer. `mu4e--server-buf gets
 its contents from the mu-servers in the following form:
    <`mu4e--server-cookie-pre'><length-in-hex><`mu4e--server-cookie-post'>
@@ -202,6 +251,29 @@ removed."
             (setq mu4e--server-buf (substring mu4e--server-buf sexp-len))
             (car objcons)))))))
 
+(defun mu4e--server-plist-get (plist key)
+  "Like `plist-get' but load data from file if it is a string.
+
+I.e. (mu4e--server-plist-get (:foo bar) :foo)
+  => bar
+but
+     (mu4e--server-plist-get (:foo \"/tmp/data.eld\") :foo)
+  => evaluates the contents of /tmp/data.eld
+   (and deletes the file afterward).
+
+This for the few sexps we get from the mu server that support this
+(headers, contacts, maildirs)."
+  ;; XXX: perhaps re-use the same buffer?
+  (let ((val (plist-get plist key)))
+    (if (stringp val)
+        (with-temp-buffer
+          (insert-file-contents val)
+          (goto-char (point-min))
+          (delete-file val)
+          (read (current-buffer)))
+      val)))
+
+
 (defun mu4e--server-filter (_proc str)
   "Filter string STR from PROC.
 This processes the \"mu server\" output. It accumulates the
@@ -269,9 +341,10 @@ The server output is as follows:
       (while sexp
         (mu4e-log 'from-server "%s" sexp)
         (cond
-         ;; a header plist can be recognized by the existence of a :date field
+         ;; a list of messages (after a find command)
          ((plist-get sexp :headers)
-          (funcall mu4e-headers-append-func (plist-get sexp :headers)))
+          (funcall mu4e-headers-append-func
+                   (mu4e--server-plist-get sexp :headers)))
 
          ;; the found sexp, we receive after getting all the headers
          ((plist-get sexp :found)
@@ -293,14 +366,19 @@ The server output is as follows:
 
          ;; received a pong message
          ((plist-get sexp :pong)
-	  (setq mu4e--server-props (plist-get sexp :props))
+          (setq mu4e--server-props (plist-get sexp :props))
           (funcall mu4e-pong-func sexp))
 
+         ;; receive queries info
+         ((plist-get sexp :queries)
+          (setq mu4e--server-query-items (plist-get sexp :queries))
+          (funcall mu4e-queries-func sexp))
+
          ;; received a contacts message
          ;; note: we use 'member', to match (:contacts nil)
          ((plist-member sexp :contacts)
           (funcall mu4e-contacts-func
-                   (plist-get sexp :contacts)
+                   (mu4e--server-plist-get sexp :contacts)
                    (plist-get sexp :tstamp)))
 
          ;; something got moved/flags changed
@@ -314,17 +392,14 @@ The server output is as follows:
          ((plist-get sexp :remove)
           (funcall mu4e-remove-func (plist-get sexp :remove)))
 
-         ;; start composing a new message
-         ((plist-get sexp :compose)
-          (funcall mu4e-compose-func
-                   (plist-get sexp :compose)
-                   (plist-get sexp :original)
-                   (plist-get sexp :include)))
-
          ;; get some info
          ((plist-get sexp :info)
           (funcall mu4e-info-func sexp))
 
+         ;; get some data
+         ((plist-get sexp :maildirs)
+          (setq mu4e-maildir-list (mu4e--server-plist-get sexp :maildirs)))
+
          ;; receive an error
          ((plist-get sexp :error)
           (funcall mu4e-error-func
@@ -341,39 +416,70 @@ As per issue #2198."
   (seq-each
    (lambda(proc)
      (when (and (process-live-p proc)
-		(string-prefix-p mu4e--server-name (process-name proc)))
+                (string-prefix-p mu4e--server-name (process-name proc)))
        (mu4e-message "killing stale mu4e server")
        (ignore-errors
-	 (signal-process proc 'SIGINT) ;; nicely
-	 (sit-for 1.0)
-	 (signal-process proc 'SIGKILL)))) ;; forcefully
+         (signal-process proc 'SIGINT) ;; nicely
+         (sit-for 1.0)
+         (signal-process proc 'SIGKILL)))) ;; forcefully
    (process-list)))
 
+(defun mu4e--server-args()
+  "Return the command line args for the command  to start the mu4e-server."
+  ;; [--debug] server [--muhome=..]
+  (seq-filter #'identity ;; filter out nil
+              `(,(when mu4e-mu-debug "--debug")
+                "server"
+                ,(when mu4e-mu-allow-temp-file "--allow-temp-file")
+                ,(when mu4e-mu-home (format "--muhome=%s" mu4e-mu-home)))))
+
+(defun mu4e--version-check ()
+  ;; sanity-check 1
+  (let ((default-directory temporary-file-directory)) ;;ensure it's local.
+    (unless (and mu4e-mu-binary (file-executable-p mu4e-mu-binary))
+      (mu4e-error
+       "Cannot find mu, please set `mu4e-mu-binary' to the mu executable path"))
+    ;; sanity-check 2
+    (let ((version (let ((s (shell-command-to-string
+                             (concat mu4e-mu-binary " --version"))))
+                     (and (string-match "version \\([.0-9]+\\)" s)
+                          (match-string 1 s)))))
+      (if (not (string= version mu4e-mu-version))
+          (mu4e-error
+           (concat
+            "Found mu version %s, but mu4e needs version %s"
+            "; please set `mu4e-mu-binary' "
+            "accordingly")
+           version mu4e-mu-version)
+        (mu4e-message "Found mu version %s" version)))))
+
+(defun mu4e-server-repl ()
+  "Start a mu4e-server repl.
+
+This is meant for debugging/testing - the repl is designed for
+machines, not for humans.
+
+You cannot run the repl when mu4e is running (or vice-versa)."
+  (interactive)
+  (if (mu4e-running-p)
+      (mu4e-error "Cannot run repl when mu4e is running")
+    (progn
+      (mu4e--version-check)
+      (let ((cmd (string-join (cons mu4e-mu-binary (mu4e--server-args)) " ")))
+        (term cmd)
+        (rename-buffer "*mu4e-repl*" 'unique)
+        (message "invoked: '%s'" cmd)))))
+
 (defun mu4e--server-start ()
   "Start the mu server process."
-  (let ((default-directory temporary-file-directory)) ;;ensure it's local.
-    ;; sanity-check 1
-  (unless (and mu4e-mu-binary (file-executable-p mu4e-mu-binary))
-    (mu4e-error
-     "Cannot find mu, please set `mu4e-mu-binary' to the mu executable path"))
-  ;; sanity-check 2
-  (let ((version (let ((s (shell-command-to-string
-			   (concat mu4e-mu-binary " --version"))))
-                   (and (string-match "version \\([.0-9]+\\)" s)
-                        (match-string 1 s)))))
-    (unless (string= version mu4e-mu-version)
-      (mu4e-error
-       (concat
-        "Found mu version %s, but mu4e needs version %s"
-	"; please set `mu4e-mu-binary' "
-        "accordingly") version mu4e-mu-version)))
+  (mu4e--version-check)
   ;; kill old/stale servers, if any.
   (mu4e--kill-stale)
   (let* ((process-connection-type nil) ;; use a pipe
-         (args (when mu4e-mu-home `(,(format"--muhome=%s" mu4e-mu-home))))
-         (args (if mu4e-mu-debug (cons "--debug" args) args))
-         (args (cons "server" args)))
+         (args (mu4e--server-args)))
     (setq mu4e--server-buf "")
+    (mu4e-log 'misc "* invoking '%s' with parameters %s" mu4e-mu-binary
+              (mapconcat (lambda (arg) (format "'%s'" arg)) args " "))
     (setq mu4e--server-process (apply 'start-process
                                       mu4e--server-name mu4e--server-name
                                       mu4e-mu-binary args))
@@ -383,7 +489,7 @@ As per issue #2198."
     (set-process-query-on-exit-flag mu4e--server-process nil)
     (set-process-coding-system mu4e--server-process 'binary 'utf-8-unix)
     (set-process-filter mu4e--server-process 'mu4e--server-filter)
-    (set-process-sentinel mu4e--server-process 'mu4e--server-sentinel))))
+    (set-process-sentinel mu4e--server-process 'mu4e--server-sentinel)))
 
 (defun mu4e--server-kill ()
   "Kill the mu server process."
@@ -408,6 +514,7 @@ As per issue #2198."
 (defun mu4e--server-sentinel (proc _msg)
   "Function called when the server process PROC terminates with MSG."
   (let ((status (process-status proc)) (code (process-exit-status proc)))
+    (mu4e-log 'misc "* famous last words from server: '%s'" mu4e--server-buf)
     (setq mu4e--server-process nil)
     (setq mu4e--server-buf "") ;; clear any half-received sexps
     (cond
@@ -415,16 +522,18 @@ As per issue #2198."
       (cond
        ((or(eq code 9) (eq code 2)) (message nil))
        ;;(message "the mu server process has been stopped"))
-       (t (error (format "mu server process received signal %d" code)))))
+       (t (mu4e-error (format "server process received signal %d" code)))))
      ((eq status 'exit)
       (cond
        ((eq code 0)
         (message nil)) ;; don't do anything
+       ((eq code 11)
+        (error "schema mismatch; please re-init mu from command-line"))
        ((eq code 19)
-        (error "Database is locked by another process"))
-       (t (error "Mu server process ended with exit code %d" code))))
+        (error "mu database is locked by another process"))
+       (t (error "mu server process ended with exit code %d" code))))
      (t
-      (error "Something bad happened to the mu server process")))))
+      (error "something bad happened to the mu server process")))))
 
 (defun mu4e--server-call-mu (form)
   "Call the mu server with some command FORM."
@@ -440,20 +549,6 @@ On success, we receive `'(:info add :path <path> :docid <docid>)'
 as well as `'(:update <msg-sexp>)`'; otherwise, we receive an error."
   (mu4e--server-call-mu `(add :path ,path)))
 
-(defun mu4e--server-compose (type decrypt &optional docid)
-  "Compose a message of TYPE, DECRYPT it and use DOCID.
-TYPE is a symbol, either `forward', `reply', `edit', `resend' or
-`new', based on an original message (ie, replying to, forwarding,
-editing, resending) with DOCID or nil for type `new'.
-
-The result is delivered to the function registered as
-`mu4e-compose-func'."
-  (mu4e--server-call-mu
-   `(compose
-     :type ,type
-     :decrypt ,(and decrypt t)
-     :docid   ,docid)))
-
 (defun mu4e--server-contacts (personal after maxnum tstamp)
   "Ask for contacts with PERSONAL AFTER MAXNUM TSTAMP.
 
@@ -470,8 +565,14 @@ get at most MAX contacts."
      :tstamp   ,(or tstamp nil)
      :maxnum   ,(or maxnum nil))))
 
+(defun mu4e--server-data (kind)
+  "Request data of some KIND.
+KIND is a symbol. Currently supported kinds: maildirs."
+  (mu4e--server-call-mu
+   `(data :kind ,kind)))
+
 (defun mu4e--server-find (query threads sortfield sortdir maxnum skip-dups
-				include-related)
+                                include-related)
   "Run QUERY with THREADS SORTFIELD SORTDIR MAXNUM SKIP-DUPS INCLUDE-RELATED.
 
 If THREADS is non-nil, show results in threaded fashion,
@@ -510,11 +611,20 @@ added or removed), since merely editing a message does not update
 the directory time stamp."
   (mu4e--server-call-mu
    `(index :cleanup ,(and cleanup t)
-	   :lazy-check ,(and lazy-check t))))
+           :lazy-check ,(and lazy-check t))))
 
-(defun mu4e--server-mkdir (path)
-  "Create a new maildir-directory at filesystem PATH."
-  (mu4e--server-call-mu `(mkdir :path ,path)))
+(defun mu4e--server-mkdir (path &optional update)
+  "Create a new maildir-directory at filesystem PATH.
+When UPDATE is non-nil, send a update when completed.
+PATH must be below the root-maildir."
+  ;; handle maildir cache
+  (if (not (string-prefix-p (mu4e-root-maildir) path))
+      (mu4e-error "Cannot create maildir outside root-maildir")
+    (add-to-list 'mu4e-maildir-list ;; update cache
+                 (substring path (length (mu4e-root-maildir)))))
+  (mu4e--server-call-mu `(mkdir
+                          :path ,path
+                          :update ,(or update nil))))
 
 (defun mu4e--server-move (docid-or-msgid &optional maildir flags no-view)
   "Move message identified by DOCID-OR-MSGID.
@@ -552,8 +662,9 @@ Returns either (:update ... ) or (:error ) sexp, which are handled my
   (unless (or maildir flags)
     (mu4e-error "At least one of maildir and flags must be specified"))
   (unless (or (not maildir)
-              (file-exists-p (concat (mu4e-root-maildir) "/" maildir "/")))
-    (mu4e-error "Target dir does not exist"))
+              (file-exists-p
+               (mu4e-join-paths (mu4e-root-maildir) maildir)))
+    (mu4e-error "Target directory does not exist"))
   (mu4e--server-call-mu
    `(move
      :docid ,(if (stringp docid-or-msgid) nil docid-or-msgid)
@@ -563,24 +674,23 @@ Returns either (:update ... ) or (:error ) sexp, which are handled my
      :rename  ,(and maildir mu4e-change-filenames-when-moving t)
      :no-view ,(and no-view t))))
 
-(defun mu4e--server-ping (&optional queries)
-  "Sends a ping to the mu server, expecting a (:pong ...) in response.
+(defun mu4e--server-ping ()
+  "Sends a ping to the mu server, expecting a (:pong ...) in response."
+  (mu4e--server-call-mu `(ping)))
+
+(defun mu4e--server-queries (queries)
+  "Sends queries to the mu server, expecting a (:queries ...) sexp in response.
 QUERIES is a list of queries for the number of results with
 read/unread status are returned in the pong-response."
-  (mu4e--server-call-mu `(ping :queries ,queries)))
+  (mu4e--server-call-mu `(queries :queries ,queries)))
 
-(defun mu4e--server-remove (docid)
-  "Remove message  with DOCID.
-The results are reporter through either (:update ... )
-or (:error) sexp, which are handled my `mu4e-error-func',
-respectively."
-  (mu4e--server-call-mu `(remove :docid ,docid)))
-
-(defun mu4e--server-sent (path)
-  "Tell the mu server we sent a message at PATH.
-If this works, we will receive (:info add :path <path> :docid
-<docid> :fcc <path>)."
-  (mu4e--server-call-mu `(sent :path ,path)))
+(defun mu4e--server-remove (docid-or-path)
+  "Remove message with either DOCID or PATH.
+The results are reported through either (:update ... )
+or (:error) sexps."
+  (if (stringp docid-or-path)
+      (mu4e--server-call-mu `(remove :path ,docid-or-path))
+      (mu4e--server-call-mu `(remove :docid ,docid-or-path))))
 
 (defun mu4e--server-view (docid-or-msgid &optional mark-as-read)
   "View a message referred to by DOCID-OR-MSGID.
